#pragma implementation

#include <debtags/Package.h>
#include <debtags/Maintainer.h>
#include <debtags/Environment.h>
#include <debtags/PackageDB.h>
#include <debtags/TagDB.h>

#include <tagcoll/TagSet.h>
#include <tagcoll/StringParserInput.h>

using namespace std;

namespace Debtags {

class PackageImpl
{
protected:
	int _ref;
	
	std::string _name;
	std::string _sdesc;
	std::string _ldesc;
	std::string _fulldata;
	Maintainer _maint;
	
public:
	PackageImpl(const std::string& name) throw ();

	/// Increment the reference count for this object
	void ref() throw () { ++_ref; }

	/// Decrement the reference count for this object, returning true when it
	/// reaches 0
	bool unref() throw () { return --_ref == 0; }

	// Fills missing fields of this package from the APT database
	void fillData() throw ();

	friend class Package;
	friend class PackageFiller;
};

};

using namespace Debtags;

PackageImpl::PackageImpl(const std::string& name) throw ()
	: _name(name)//, _maint(0)
{
	/*
	int send = desc.find('\n');
	_desc = desc.substr(0, send);
	_ldesc = desc.substr(send + 1);
	*/
}

void PackageImpl::fillData() throw ()
{
	//fprintf(stderr, "FILLDATA %.*s\n", PFSTR(_name));
	Package pkg(this);
	Debtags::Environment::get().fillPackageData(pkg);
/*
	fprintf(stderr, "AFTER FILL:\n");
	fprintf(stderr, "  name: %.*s\n", PFSTR(_name));
	fprintf(stderr, "  sdesc: %.*s\n", PFSTR(_sdesc));
	fprintf(stderr, "  ldesc: %.*s\n", PFSTR(_ldesc));
	fprintf(stderr, "  fulldata: %.*s\n", PFSTR(_fulldata));
	fprintf(stderr, "  maint: %s\n", _maint ? "present" : "null");
*/
}



Package::Package() throw ()
	: impl(0) {}

Package::Package(const std::string& name) throw ()
{
	impl = new PackageImpl(name);
	impl->ref();
}

Package::Package(PackageImpl* existing_impl) throw ()
{
	if (existing_impl)
		existing_impl->ref();
	impl = existing_impl;
}

Package::Package(const Package& ns) throw ()
{
	if (ns.impl)
		ns.impl->ref();
	impl = ns.impl;
}

Package::~Package() throw ()
{   
	if (impl && impl->unref())
		delete impl;
}

Package& Package::operator=(const Package& ns) throw ()
{   
	if (ns.impl)
		ns.impl->ref();  // Do it early to correctly handle the case of x = x;
	if (impl && impl->unref())
		delete impl;
	impl = ns.impl;
	return *this;
}

bool Package::operator==(const Package& ns) const throw ()
{   
	return name() == ns.name();
}

bool Package::operator<(const Package& ns) const throw ()
{   
	return name() < ns.name();
}

Package::operator bool() const throw ()
{
	return impl != 0;
}

const std::string& Package::name() const throw () { return impl->_name; }
const Maintainer& Package::maint() const throw ()
{
	if (! (bool)impl->_maint)
		impl->fillData();
	return impl->_maint;
}
const std::string& Package::sdesc() const throw ()
{
	if (impl->_sdesc.empty())
		impl->fillData();
	return impl->_sdesc;
}
const std::string& Package::ldesc() const throw ()
{
	if (impl->_ldesc.empty())
		impl->fillData();
	return impl->_ldesc;
}
const std::string& Package::fulldata() const throw ()
{
	if (impl->_fulldata.empty())
		impl->fillData();
	return impl->_fulldata;
}

int Package::state() const throw (PackageDBException)
{
	return Environment::get().packageDB().getState(name());
}

const TagSet Package::tags() const throw ()
{
	return Environment::get().tagDB().getTagsetForItem(*this);
}

const std::string& Package::sdesc(const std::string& sdesc) throw ()
{
	return impl->_sdesc = sdesc;
}
const std::string& Package::ldesc(const std::string& ldesc) throw ()
{
	return impl->_ldesc = ldesc;
}
const Maintainer& Package::maint(const Maintainer& maint) throw ()
{
	return impl->_maint = maint;
}
const std::string& Package::fulldata(const std::string& fulldata) throw ()
{
	return impl->_fulldata = fulldata;
}

// vim:set ts=3 sw=3:
