/*
 * Wrapper around libapt-pkg
 *
 * Copyright (C) 2003  Enrico Zini <enrico@debian.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * FIXME: this code contain lots of guesses about the real functioning of
 * apt-pkg, mainly taken by just pasting code from apt-cache.  If someone knows
 * about some real documentation for libapt-pkg API, please fix this code.
 *
 * I wrote "please fix this code" and not "please tell me" because there's too
 * much infrastructure in libapt-pkg for me to learn just to fetch out some
 * package description.
 *
 * If I may, I'd like to suggest that someone knowledgeable about libapt-pkg
 * could write some simplified kind of access to libapt-pkg package cache.
 *
 * On my part, I'm doing my best to isolate the rest of the code from
 * libapt-pkg interface code: since I don't know it that well, I try to keep
 * mistakes contained in this files.
 */

#include <debtags/PackageDB.h>

#include <list>
#include <queue>

#include <apt-pkg/error.h>
#include <apt-pkg/pkgcachegen.h>
#include <apt-pkg/init.h>
#include <apt-pkg/progress.h>
#include <apt-pkg/sourcelist.h>
#include <apt-pkg/cmndline.h>
#include <apt-pkg/strutl.h>
#include <apt-pkg/pkgrecords.h>
#include <apt-pkg/srcrecords.h>
#include <apt-pkg/version.h>
#include <apt-pkg/policy.h>
#include <apt-pkg/tagfile.h>
#include <apt-pkg/algorithms.h>
#include <apt-pkg/sptr.h>
#include <apt-pkg/tagfile.h>

using namespace std;
using namespace Debtags;

namespace Debtags
{
	
class PackageDBImpl
{
protected:
	int _ref;

	pkgSourceList *SrcList;
	MMap *Map;
	pkgCache *GCache;

public:
	PackageDBImpl() throw (PackageDBException);
	~PackageDBImpl() throw ();

	/// Increment the reference count for this object
	void ref() throw () { ++_ref; }

	/// Decrement the reference count for this object, returning true when it
	/// reaches 0
	bool unref() throw () { return --_ref == 0; }

	bool hasPackage(const string& pkg) const throw ();
	string getPackageRecord(const string& pkg) const throw (PackageDBException);
	void getPackageRecord(const std::string& pkg, pkgTagSection& section) const
		throw (PackageDBException);

	int getState(const string& pkg) const throw (PackageDBException);

	void iterateAll(PkgRecordConsumer& consume) const throw (PackageDBException);
};

};

static void throwPackageDBException(const string& context) throw (PackageDBException)
{
	string err;
	string errs;
	bool first = true;
	while (_GetErrorObj()->PendingError())
	{
		_GetErrorObj()->PopMessage(err);
		if (first)
		{
			errs += err;
			first = false;
		}
		else
			errs += " - " + err;
	}
	throw PackageDBException(errs + " " + context);
}

PackageDBImpl::PackageDBImpl() throw (PackageDBException)
	: _ref(0), SrcList(0), Map(0), GCache(0)
{
	_config->Set("quiet",0);
	_config->Set("help",false); 

	if (pkgInitConfig(*_config) == false ||
			pkgInitSystem(*_config,_system) == false)
		throwPackageDBException("initializing apt-pkg library");

	// Open the cache file
	SrcList = new pkgSourceList;
	SrcList->ReadMainList();

	// Generate it and map it
	OpProgress Prog;
	pkgMakeStatusCache(*SrcList,Prog,&Map,true);

	if (_error->PendingError() == true)
		throwPackageDBException("after pkgMakeStatusCache");

	GCache = new pkgCache(Map);

	if (_error->PendingError() == true)
		throwPackageDBException("after creating new Cache");
}

PackageDBImpl::~PackageDBImpl() throw ()
{
	delete GCache;
	delete Map;
	delete SrcList;
}

bool PackageDBImpl::hasPackage(const string& pkg) const throw ()
{
	pkgCache::PkgIterator Pkg = GCache->FindPkg(pkg);
	return Pkg.end() != true;
}


// LocalitySort - Sort a version list by package file locality
static int LocalityCompare(const void *a, const void *b)
{
   pkgCache::VerFile *A = *(pkgCache::VerFile **)a;
   pkgCache::VerFile *B = *(pkgCache::VerFile **)b;
   
   if (A == 0 && B == 0)
      return 0;
   if (A == 0)
      return 1;
   if (B == 0)
      return -1;
   
   if (A->File == B->File)
      return A->Offset - B->Offset;
   return A->File - B->File;
}
static void LocalitySort(pkgCache::VerFile **begin,
		  unsigned long Count,size_t Size)
{   
   qsort(begin,Count,Size,LocalityCompare);
}

// taken from apt-cache.cc in apt
void PackageDBImpl::iterateAll(PkgRecordConsumer& consume) const throw (PackageDBException)
{
	pkgCache &Cache = *GCache;
	pkgDepCache::Policy Plcy;

	// Create the text record parser
	pkgRecords Recs(Cache);
	if (_error->PendingError() == true)
		throwPackageDBException("can't create a record parser [FIXME: Enrico has no idea on how to handle this situation]");

	pkgCache::VerFile **VFList = new pkgCache::VerFile*[Cache.HeaderP->PackageCount+1];
	memset(VFList,0,sizeof(*VFList)*Cache.HeaderP->PackageCount+1);

	// Map versions that we want to write out onto the VerList array.
	for (pkgCache::PkgIterator P = Cache.PkgBegin(); P.end() == false; P++)
	{
		// Find the proper version to use. 
		pkgCache::VerIterator V = Plcy.GetCandidateVer(P);
		if (V.end() == false)
			VFList[P->ID] = V.FileList();
	}

	// Include all the packages that provide matching names too
	for (pkgCache::PkgIterator P = Cache.PkgBegin(); P.end() == false; P++)
		for (pkgCache::PrvIterator Prv = P.ProvidesList() ; Prv.end() == false; Prv++)
		{
			pkgCache::VerIterator V = Plcy.GetCandidateVer(Prv.OwnerPkg());
			if (V.end() == false)
				VFList[Prv.OwnerPkg()->ID] = V.FileList();
		}

	// Sort things by file and offset, so that the next data read loop is more
	// efficient
	LocalitySort(VFList,Cache.HeaderP->PackageCount,sizeof(*VFList));

	// Iterate over all the version records and check them
	for (pkgCache::VerFile **J = VFList; *J != 0; J++)
	{
		pkgRecords::Parser &P = Recs.Lookup(pkgCache::VerFileIterator(Cache,*J));

		consume(P);
	}

	delete [] VFList;
}


void PackageDBImpl::getPackageRecord(const std::string& pkg, pkgTagSection& section) const
	throw (PackageDBException)
{
	if (_error->PendingError())
	{
		fprintf(stderr, "PackageDBImpl::getPackageRecord: Discarding old errors...\n");
		_error->Discard();
	}
	
	pkgCache::PkgIterator Pkg = GCache->FindPkg(pkg);
	if (Pkg.end() == true)
		throw PackageDBException("Unable to locate package \"" + pkg + "\"");

	// Find the proper version to use.
	pkgDepCache::Policy Plcy;

	pkgCache::VerIterator V = Plcy.GetCandidateVer(Pkg);
	if (V.end() == true || V.FileList().end() == true)
		throw PackageDBException("Can't find a candidate version for package \"" + pkg + "\" [FIXME: this error message text is just an uninformed guess by Enrico about what might have gone wrong]");

	// Find an appropriate file
	pkgCache::VerFileIterator Vf = V.FileList();
	for (; Vf.end() == false; Vf++)
		if ((Vf.File()->Flags & pkgCache::Flag::NotSource) == 0)
			break;
	if (Vf.end() == true)
		Vf = V.FileList();

	// Check and load the package list file
	pkgCache::PkgFileIterator I = Vf.File();
	if (I.IsOk() == false)
		throw PackageDBException(string("Package file ") + I.FileName() + " is out of sync");

	FileFd PkgF(I.FileName(),FileFd::ReadOnly);
	if (_error->PendingError() == true)
		throwPackageDBException("while reading a package data file [FIXME: Enrico has no idea on how to handle this situation]");

	pkgTagFile file(&PkgF);
	if (_error->PendingError() == true)
		throwPackageDBException("can't create a pkgTagFile object to access data file [FIXME: Enrico has no idea on how to handle this situation]");

	if (!file.Jump(section, V.FileList()->Offset))
		throwPackageDBException(string("can't create read record from data file ") + I.FileName() + " [FIXME: Enrico has no idea on how to handle this situation]");
}

string PackageDBImpl::getPackageRecord(const string& pkg) const
	throw (PackageDBException)
{
	if (_error->PendingError())
	{
		fprintf(stderr, "PackageDBImpl::getPackageRecord: Discarding old errors...\n");
		_error->Discard();
	}
	
	pkgCache::PkgIterator Pkg = GCache->FindPkg(pkg);
	if (Pkg.end() == true)
		throw PackageDBException("Unable to locate package \"" + pkg + "\"");

	// Find the proper version to use.
	pkgDepCache::Policy Plcy;

	pkgCache::VerIterator V = Plcy.GetCandidateVer(Pkg);
	if (V.end() == true || V.FileList().end() == true)
		throw PackageDBException("Can't find a candidate version for package \"" + pkg + "\" [FIXME: this error message text is just an uninformed guess by Enrico about what might have gone wrong]");

	// Find an appropriate file
	pkgCache::VerFileIterator Vf = V.FileList();
	for (; Vf.end() == false; Vf++)
		if ((Vf.File()->Flags & pkgCache::Flag::NotSource) == 0)
			break;
	if (Vf.end() == true)
		Vf = V.FileList();

	// Check and load the package list file
	pkgCache::PkgFileIterator I = Vf.File();
	if (I.IsOk() == false)
		throw PackageDBException(string("Package file ") + I.FileName() + " is out of sync");

	FileFd PkgF(I.FileName(),FileFd::ReadOnly);
	if (_error->PendingError() == true)
		throwPackageDBException("while reading a package data file [FIXME: this error message text is just an uninformed guess by Enrico about what might have gone wrong]");

	// Read the record and then write it out again.
	char *Buffer = new char[GCache->HeaderP->MaxVerFileSize + 1];
	Buffer[V.FileList()->Size] = '\n';
	if (PkgF.Seek(V.FileList()->Offset) == false ||
			PkgF.Read(Buffer,V.FileList()->Size) == false)
	{
		delete [] Buffer;
		throw PackageDBException("Can't read package data for package \"" + pkg + "\" [FIXME: this error message text is just an uninformed guess by Enrico about what might have gone wrong]");
	}

	string res(Buffer, V.FileList()->Size);
	delete [] Buffer;
	return res;
}

int PackageDBImpl::getState(const string& pkg) const throw (PackageDBException)
{
	pkgCache::PkgIterator Pkg = GCache->FindPkg(pkg);
	if (Pkg.end() == true)
		throw PackageDBException("Unable to locate package \"" + pkg + "\"");
	//printf("IS: %c %d\n", Pkg->CurrentState, (int)Pkg->CurrentState);
	return Pkg->CurrentState;
}

PackageDB::PackageDB() throw (PackageDBException)
{
	impl = new PackageDBImpl();
	impl->ref();
}

PackageDB::PackageDB(const PackageDB& ns) throw ()
{
	if (ns.impl)
		ns.impl->ref();
	impl = ns.impl;
}

PackageDB::~PackageDB() throw ()
{
	if (impl && impl->unref())
		delete impl;
}

PackageDB& PackageDB::operator=(const PackageDB& ns) throw ()
{
	if (ns.impl)
		ns.impl->ref();  // Do it early to correctly handle the case of x = x;
	if (impl && impl->unref())
		delete impl;
	impl = ns.impl;
	return *this;
}

bool PackageDB::hasPackage(const std::string& pkg) const throw ()
{
	return impl->hasPackage(pkg);
}

std::string PackageDB::getPackageRecord(const std::string& pkg) const throw (PackageDBException)
{
	return impl->getPackageRecord(pkg);
}

void PackageDB::getPackageRecord(const std::string& pkg, pkgTagSection& section) const
	throw (PackageDBException)
{
	impl->getPackageRecord(pkg, section);
}

int PackageDB::getState(const std::string& pkg) const throw (PackageDBException)
{
	return impl->getState(pkg);
}

void PackageDB::iterateAll(PkgRecordConsumer& consume) const throw (PackageDBException)
{
	impl->iterateAll(consume);
}

// vim:set ts=4 sw=4:
