/*
 * Central class to perform package searches
 *
 * Copyright (C) 2003  Enrico Zini <enrico@debian.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#pragma implementation

#include <debtags/PackageSet.h>
#include <debtags/MaintainerSet.h>
#include <debtags/Environment.h>
#include <tagcoll/TagSet.h>

using namespace std;
using namespace Debtags;
using namespace Tagcoll;


PackageSet::PackageSet() throw ()
{
}

PackageSet::PackageSet(const OpSet<Debtags::Package>& p) throw ()
	: OpSet<Debtags::Package>(p)
{
}

PackageSet::~PackageSet() throw ()
{
}

void PackageSet::consume(const Package& pkg) throw ()
{
	*this += pkg;
}


FacetSet PackageSet::getFacets() const throw ()
{
	FacetSet res;
	for (const_iterator i = begin(); i != end(); i++)
		res += i->tags().getFacets();
	return res;
}

TagSet PackageSet::getTags() const throw ()
{
	TagSet res;
	for (const_iterator i = begin(); i != end(); i++)
		res += i->tags();
	return res;
}

MaintainerSet PackageSet::getMaintainers() const throw ()
{
	MaintainerSet res;
	for (const_iterator i = begin(); i != end(); i++)
		res += i->maint();
	return res;
}

// Get the set of all tags in this collection that appear in tagsets
// containing `ts'
TagSet PackageSet::getCompanionTags(const TagSet& ts) const throw ()
{
	TagSet tags;

	for (const_iterator i = begin(); i != end(); i++)
		if (i->tags().contains(ts))
			tags += i->tags() - ts;

	return tags;
}

Package PackageSet::find(const std::string& name) const throw ()
{
	Package sample(name);
	const_iterator found = find(sample);
	if (found != end())
		return *found;
	else
		return Package();
}

Maintainer PackageSet::findMaintainer(const std::string& email) const throw ()
{
	for (const_iterator i = begin(); i != end(); i++)
		if (i->maint().email() == email)
			return i->maint();
	return Maintainer();
}

void PackageSet::output(PackageConsumer& cons) const throw ()
{
	for (iterator i = begin(); i != end(); i++)
		cons.consume(*i);
}

void PackageSet::output(PackageConsumer& cons, PackageMatcher& filter) const throw ()
{
	for (iterator i = begin(); i != end(); i++)
		if (filter.match(*i))
			cons.consume(*i);
}

PackageSet PackageSet::getFiltered(PackageMatcher& filter) const throw ()
{
	PackageSet res;
	for (iterator i = begin(); i != end(); i++)
		if (filter.match(*i))
			res += *i;
	return res;
}

// Get the PackageSet of packages related to `pivot'
PackageSet PackageSet::getRelated(const Package& pivot, int distance) const throw ()
{
	return getRelated(pivot.tags(), distance) - pivot;
}

// Get the PackageSet of packages related to the specific tagset
PackageSet PackageSet::getRelated(const TagSet& ts, int distance) const throw ()
{
	PackageSet res;

	for (const_iterator i = begin(); i != end(); i++)
	{
		int dist = ts.distance(i->tags());
		if (dist >= 0 && dist <= distance)
			res += *i;
	}

	return res;

}

void PackageSet::outputRelated(PackageConsumer& cons, const Package& pivot, int distance) const throw ()
{
	PackageSet res = getRelated(pivot, distance);
	//fprintf(stderr, "%d related to %.*s, distance %d\n", res.size(), PFSTR(pivot.name()), distance);
	for (PackageSet::const_iterator i = res.begin(); i != res.end(); i++)
		cons.consume(*i);
}

void PackageSet::outputRelated(PackageConsumer& cons, const TagSet& ts, int distance) const throw ()
{
	PackageSet res = getRelated(ts, distance);
	for (PackageSet::const_iterator i = res.begin(); i != res.end(); i++)
		cons.consume(*i);
}

// vim:set ts=4 sw=4:
