/*
 * System tag database
 *
 * Copyright (C) 2003  Enrico Zini <enrico@debian.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#pragma implementation

#include <debtags/TagDB.h>
#include <debtags/Vocabulary.h>
#include <debtags/Environment.h>
#include <debtags/Package.h>

#include <tagcoll/StdioParserInput.h>
#include <tagcoll/TagcollParser.h>
#include <tagcoll/TagcollFilter.h>
#include <tagcoll/DerivedTags.h>
#include <tagcoll/FilterChain.h>
#include <tagcoll/Implications.h>
#include <tagcoll/Patches.h>
#include <tagcoll/TagcollSerializer.h>
#include <tagcoll/UnfacetedRemover.h>

#include <sys/wait.h>	// WIFEXITED WEXITSTATUS
#include <sys/types.h>	// getpwuid, stat, mkdir, getuid
#include <sys/stat.h>	// stat, mkdir
#include <pwd.h>		// getpwuid
#include <unistd.h>		// stat, getuid

#include <errno.h>

using namespace std;
using namespace Debtags;

static const std::string _path_tagdb = STATEDIR "/package-tags";
static const std::string _path_vocab = STATEDIR "/vocabulary";
static std::string rcdir;

/*
namespace {
static string fmtts(const OpSet<string>& tags) throw ()
{
	string res;
	for (OpSet<string>::const_iterator i = tags.begin();
			i != tags.end(); i++)
		if (i == tags.begin())
			res += *i;
		else
			res += ", " + *i;
	return res;
}

template<class ITEM, class TAG = std::string>
class TagSniffer : public TagcollFilter<ITEM, TAG>
{
protected:
	string prefix;

public:
	TagSniffer(const string prefix) throw () : prefix(prefix) {}
	virtual ~TagSniffer() throw() {}
//	TagSniffer(TagcollConsumer<string, string>& target) throw ()
//		: target(target) {}

	virtual void consume(const std::string& item) throw ()
	{
		if (item == "hdparm")
			printf("%.*s Single no ts: hdparm\n", PFSTR(prefix));
		consumer->consume(item);
	}
	virtual void consume(const std::string& item, const OpSet<std::string>& tags) throw ()
	{
		if (item == "hdparm")
		{
			string s = fmtts(tags);
			printf("%.*s Single: hdparm (%.*s)\n", PFSTR(prefix), PFSTR(s));
		}
		consumer->consume(item, tags);
	}
	virtual void consume(const OpSet<ITEM>& items) throw ()
	{
		if (items.find("hdparm") != items.end())
			printf("%.*s Multi no ts: hdparm\n", PFSTR(prefix));
		consumer->consume(items);
	}
	virtual void consume(const OpSet<ITEM>& items, const OpSet<TAG>& tags) throw ()
	{
		if (items.find("hdparm") != items.end())
		{
			string s = fmtts(tags);
			printf("%.*s Multi: hdparm (%.*s)\n", PFSTR(prefix), PFSTR(s));
		}
		consumer->consume(items, tags);
	}
};
};
*/

static void init_rcdir(bool editable) throw (SystemException, ConsistencyCheckException)
{
	struct passwd* udata = getpwuid(getuid());
	rcdir = udata->pw_dir;
	rcdir += "/.debtags";

	//printf("Init debtags environment; rcdir is %.*s\n", PFSTR(rcdir));

	struct stat rcdir_stat;
	if (stat(rcdir.c_str(), &rcdir_stat) == -1)
	{
		//printf("Creating profile directory %.*s.\n", PFSTR(rcdir));
		if (editable && mkdir(rcdir.c_str(), 0777) == -1)
			throw SystemException(errno, "creating directory " + rcdir);
	} else {
		if (S_ISDIR(rcdir_stat.st_mode) == 0)
			if (editable)
				throw ConsistencyCheckException(rcdir + " already exists and is not a directory");
			else
				fprintf(stderr, "warning: %.*s already exists and is not a directory", PFSTR(rcdir));
		
		if (access(rcdir.c_str(), R_OK | W_OK | X_OK) == -1)
			if (editable)
				throw ConsistencyCheckException(rcdir + " already exists and has insufficient permissions");
			else
				fprintf(stderr, "%.*s already exists and has insufficient permissions", PFSTR(rcdir));
	}
}

static void readCollection(const string& file, TagcollConsumer<string>& consumer)
	throw (FileException, ParserException)
{
	StdioParserInput in(file);
	Tagcoll::parseTagcoll(in, consumer);
	/*
	TagSniffer<string, string> s("READCOLL");
	s.setConsumer(&consumer);
	Tagcoll::parseTagcoll(in, s);
	*/
}

namespace Debtags
{
// TagcollFilter that adds a "facet-only" tag for every faceted tag
// For example, if a package is tagged use::chatting, it will become tagged
// with "use, use::chatting"
template<class ITEM, class TAG = std::string>
class FacetTagAdder : public TagcollFilter<ITEM, TAG>
{
public:
	virtual void consume(const std::string& item) throw ();
	virtual void consume(const std::string& item, const OpSet<std::string>& tags) throw ();
	virtual void consume(const OpSet<ITEM>& items) throw ();
	virtual void consume(const OpSet<ITEM>& items, const OpSet<TAG>& tags) throw ();
};

class StringsToTags : public TagcollConsumer<string, string>
{
protected:
	TagcollConsumer<std::string, Debtags::Tag>& target;
	Debtags::Vocabulary& voc;
	
public:
	StringsToTags(TagcollConsumer<std::string, Debtags::Tag>& target) throw ()
		: target(target), voc(Debtags::Environment::get().vocabulary()) {}

	virtual void consume(const string& item) throw ()
	{
		target.consume(item);
	}
	virtual void consume(const string& item, const OpSet<string>& tags) throw ()
	{
		TagSet outTags;
		for (OpSet<string>::const_iterator i = tags.begin(); i != tags.end(); i++)
			if (Tag t = voc.findTag(*i))
				outTags += t;
		target.consume(item, outTags);
	}
};

class ItemsToPackages : public TagcollConsumer<string, string>
{
protected:
	TagcollConsumer<Debtags::Package, Debtags::Tag>& target;
	Debtags::Vocabulary& voc;
	
public:
	ItemsToPackages(TagcollConsumer<Debtags::Package, Debtags::Tag>& target) throw ()
		: target(target), voc(Debtags::Environment::get().vocabulary()) {}

	virtual void consume(const string& item) throw ()
	{
		/*
		if (item == "hdparm")
			printf("ITP Single no ts: hdparm\n");
		*/

		Package pkg = Environment::get().getPackage(item);
		if (!pkg)
			fprintf(stderr, "ItemsToPackages::consume(item): Name %.*s gave null Package: ignoring it.\n", PFSTR(item));
		else
			target.consume(pkg);
	}
	virtual void consume(const string& item, const OpSet<string>& tags) throw ()
	{
		/*
		if (item == "hdparm")
		{
			string s = fmtts(tags);
			printf("ITP IN Single: hdparm (%.*s)\n", PFSTR(s));
		}
		*/

		TagSet outTags;
		for (OpSet<string>::const_iterator i = tags.begin(); i != tags.end(); i++)
			if (Tag t = voc.findTag(*i))
				outTags += t;

		/*
		if (item == "hdparm")
		{
			OpSet<string> otags;
			for (TagSet::const_iterator i = outTags.begin(); i != outTags.end(); i++)
				otags += i->fullname();
			string s = fmtts(otags);
			printf("ITP Single: hdparm (%.*s)\n", PFSTR(s));
		}
		*/

		Package pkg = Environment::get().getPackage(item);
		if (!pkg)
			fprintf(stderr, "ItemsToPackages::consume(item, tags): Name %.*s gave null Package: ignoring it.\n", PFSTR(item));
		else
			target.consume(pkg, outTags);
	}
};

class PackagesToItems : public TagcollConsumer<Debtags::Package, Debtags::Tag>
{
protected:
	TagcollConsumer<string>& target;

public:
	PackagesToItems(TagcollConsumer<string>& target) throw ()
		: target(target) {}

	void consume(const Debtags::Package& item) throw ()
	{
		target.consume(item.name());
	}
	void consume(const Debtags::Package& item, const OpSet<Tag>& tags) throw ()
	{
		OpSet<string> outTags;
		for (OpSet<Tag>::const_iterator i = tags.begin(); i != tags.end(); i++)
			outTags += i->fullname();
		target.consume(item.name(), outTags);
	}
};

};


template<class ITEM, class TAG>
void FacetTagAdder<ITEM, TAG>::consume(const string& item) throw ()
{
	consumer->consume(item);
}

template<class ITEM, class TAG>
void FacetTagAdder<ITEM, TAG>::consume(const string& item, const OpSet<string>& tags) throw ()
{
	OpSet<string> added;
	for (OpSet<string>::const_iterator i = tags.begin();
			i != tags.end(); i++)
	{
		size_t p = i->find("::");
		if (p != string::npos)
			added += i->substr(0, p);
	}

	if (added.size())
		consumer->consume(item, tags + added);
	else
		consumer->consume(item, tags);
}

// Process a set of items, all with no tags
template<class ITEM, class TAG>
void FacetTagAdder<ITEM, TAG>::consume(const OpSet<ITEM>& items) throw ()
{
	consumer->consume(items);
}

// Process a set of items identically tagged, with their tags
template<class ITEM, class TAG>
void FacetTagAdder<ITEM, TAG>::consume(const OpSet<ITEM>& items, const OpSet<TAG>& tags) throw ()
{
	OpSet<string> added;
	for (OpSet<string>::const_iterator i = tags.begin();
			i != tags.end(); i++)
	{
		size_t p = i->find("::");
		if (p != string::npos)
			added += i->substr(0, p);
	}

	if (added.size())
		consumer->consume(items, tags + added);
	else
		consumer->consume(items, tags);
}


TagDB::TagDB(bool editable) throw (SystemException, ConsistencyCheckException)
{
	if (rcdir.empty())
		init_rcdir(editable);
}

TagDB::~TagDB() throw ()
{
}

const std::string& TagDB::statedir() throw () { return rcdir; }
	
const std::string& TagDB::path_tagdb() throw ()
{
	return _path_tagdb;
}

bool TagDB::hasTagDatabase() throw ()
{
	if (   access(_path_tagdb.c_str(), R_OK) == -1
		|| access(_path_vocab.c_str(), R_OK) == -1)
		return false;
	return true;
}

Debtags::TagSet TagDB::getTags() const throw ()
{
	return getAllTags();
}

void TagDB::load(bool facet_only) throw (FileException, ParserException)
{
	outputPatched(*this, facet_only);
}


void TagDB::savePatch()
	throw (FileException, ParserException, SystemException)
{
	savePatch(*this);
}

void TagDB::savePatch(const InputMerger<Debtags::Package, Debtags::Tag>& collection)
	throw (FileException, ParserException, SystemException)
{
	InputMerger<string, string> system;
	outputSystem(system);

	InputMerger<string, string> changed;
	PackagesToItems pti(changed);
	collection.output(pti);

	PatchList<string> patch;
	patch.generate(system, changed);

	string patchFile = rcdir + "/patch";
	string backup = patchFile + "~";

	if (access(patchFile.c_str(), F_OK) == 0)
		if (rename(patchFile.c_str(), backup.c_str()) == -1)
			throw SystemException(errno, "Can't rename " + patchFile + " to " + backup);

	try {
		FILE* out = fopen(patchFile.c_str(), "w");
		if (out == 0)
			throw SystemException(errno, "Can't write to " + patchFile);
		
		TagcollSerializer output(out);
		patch.output(output);

		fclose(out);
	} catch (Exception& e) {
		if (rename(backup.c_str(), patchFile.c_str()) == -1)
			fprintf(stderr, "Warning: Impossible to restore previous backup copy: %.*s (%s)\n", PFSTR(e.desc()), e.type());
		throw;
	}
}

void TagDB::sendPatch()
	throw (FileException, ParserException, SystemException, ConsistencyCheckException)
{
	sendPatch(*this);
}

void TagDB::sendPatch(const InputMerger<Debtags::Package, Debtags::Tag>& collection)
	throw (FileException, ParserException, SystemException, ConsistencyCheckException)
{
	InputMerger<string, string> system;
	outputSystem(system);

	InputMerger<string, string> changed;
	PackagesToItems pti(changed);
	collection.output(pti);

	PatchList<string> patch;
	patch.generate(system, changed);

	static const char* cmd = "/usr/sbin/sendmail -t";
	FILE* out = popen(cmd, "w");
	if (out == 0)
		throw SystemException(errno, string("trying to run `") + cmd + "'");

	struct passwd* udata = getpwuid(getuid());

	fprintf(out,
			"To: debtags-submit@vitavonni.de\n"
			"Bcc: %s\n"
			"Subject: Tag patch\n"
			"Mime-Version: 1.0\n"
			"Content-Type: multipart/mixed; boundary=\"9amGYk9869ThD9tj\"\n"
			"Content-Disposition: inline\n"
			"X-Mailer: debtags-edit\n\n"
			"This mail contains a Debtags patch for the central archive\n\n"
			"--9amGYk9869ThD9tj\n"
			"Content-Type: text/plain; charset=utf-8\n"
			"Content-Disposition: inline\n\n"
			"-- DEBTAGS DIFF V0.1 --\n", udata->pw_name);

	TagcollSerializer output(out);
	patch.output(output);

	fprintf(out, "\n--9amGYk9869ThD9tj\n");

	int res = pclose(out);
	if (!WIFEXITED(res) || WEXITSTATUS(res) != 0)
		throw ConsistencyCheckException("sendmail returned nonzero (" + stringf::fmt(res) + "): the mail may have not been sent");
}


// Output the given file to a TagcollConsumer
void TagDB::outputFile(const std::string& file, TagcollConsumer<std::string, std::string>& cons, bool facet_only)
	throw (FileException, ParserException)
{
	// Prepare the input filter chain
	FilterChain<string, string> filters;

	/*
	TagSniffer<string> of1("OF1");
	filters.appendFilter(&of1);

	TagSniffer<string> of2("OF2");
	TagSniffer<string> of3("OF3");
	*/

	/*
	ApplyDerivedTags<string> adtags(Debtags::Environment::get().vocabulary().getEquations());
	filters.appendFilter(&adtags);
	*/

	UnfacetedRemover<string> ufr;
	FacetTagAdder<string> fta;
	if (facet_only)
	{
		// Remove unfaceted tags
		filters.appendFilter(&ufr);
		//filters.appendFilter(&of2);
	} else {
		// Add facet-only tags
		filters.appendFilter(&fta);
		//filters.appendFilter(&of3);
	}

	// Read and expand the database
	filters.setConsumer(&cons);
	readCollection(file, filters);
}

// Output the given file to a TagcollConsumer
void TagDB::outputFile(const std::string& file, TagcollConsumer<string, Tag>& cons, bool facet_only)
	throw (FileException, ParserException)
{
	// Filter to convert tag names into Tag classes
	StringsToTags conv(cons);

	outputFile(file, conv, facet_only);
}

// Output the given file to a TagcollConsumer
void TagDB::outputFile(const std::string& file, TagcollConsumer<Package, Tag>& cons, bool facet_only)
	throw (FileException, ParserException)
{
	// Filter to convert into Debtags::Package
	ItemsToPackages itp(cons);

	outputFile(file, itp, facet_only);
}


// Output the current Debian tags database to a TagcollConsumer
void TagDB::outputSystem(TagcollConsumer<std::string, std::string>& cons, bool facet_only)
	throw (FileException, ParserException)
{
	outputFile(_path_tagdb, cons, facet_only);
}

// Output the current Debian tags database to a TagcollConsumer
void TagDB::outputSystem(TagcollConsumer<std::string, Tag>& cons, bool facet_only)
	throw (FileException, ParserException)
{
	outputFile(_path_tagdb, cons, facet_only);
}

// Output the current Debian tags database to a TagcollConsumer
void TagDB::outputSystem(TagcollConsumer<Package, Tag>& cons, bool facet_only)
	throw (FileException, ParserException)
{
	outputFile(_path_tagdb, cons, facet_only);
}


// Output the current Debian tags database, patched with local patch, to a TagcollConsumer
void TagDB::outputPatched(TagcollConsumer<std::string, std::string>& cons, bool facet_only)
	throw (FileException, ParserException)
{
	// Prepare the input filter chain
	FilterChain<string, string> filters;

	// Apply local patch if it exists
	PatchList<string> patches;
	string patchFile = rcdir + "/patch";
	if (access(patchFile.c_str(), R_OK) == 0)
	{
		readCollection(patchFile, patches);
		filters.appendFilter(new ApplyPatches<string>(patches));
	}

	/*
	// Expand implications
	ImplicationList implications;
	readCollection(fn_impls, implications);
	// Pack the structure for faster expansions
	implications.pack();
	TagcollFilter<string>* ximpl = new ApplyImplications(implications);
	filters.appendFilter(ximpl);

	// Add derived tags
	DerivedTagList derivedTags;
	readDerivedTags(fn_deriv, derivedTags);
	TagcollFilter<string>* xderv = new ApplyDerivedTags(derivedTags);
	filters.appendFilter(xderv);

	// Add further tags implicated by the derived tags
	filters.appendFilter(ximpl);
	*/

	/*
	ApplyDerivedTags<string> adtags(Debtags::Environment::get().vocabulary().getEquations());
	filters.appendFilter(&adtags);
	*/

	UnfacetedRemover<string> ufr;
	FacetTagAdder<string> fta;
	if (facet_only)
	{
		// Remove unfaceted tags
		filters.appendFilter(&ufr);
	} else {
		// Add facet-only tags
		filters.appendFilter(&fta);
	}

	// Read and expand the database
	filters.setConsumer(&cons);
	readCollection(_path_tagdb, filters);

	/*
	delete ximpl;
	delete xderv;
	*/
}

// Output the current Debian tags database, patched with local patch, to a TagcollConsumer
void TagDB::outputPatched(TagcollConsumer<std::string, Debtags::Tag>& cons, bool facet_only)
	throw (FileException, ParserException)
{
	// Filter to convert tag names into Tag classes
	StringsToTags conv(cons);

	outputPatched(conv, facet_only);
}

// Output the current Debian tags database, patched with local patch, to a TagcollConsumer
void TagDB::outputPatched(TagcollConsumer<Debtags::Package, Debtags::Tag>& cons, bool facet_only)
	throw (FileException, ParserException)
{
	// Convert into Debtags::Package
	ItemsToPackages itp(cons);

	outputPatched(itp, facet_only);
}

// vim:set ts=4 sw=4:
