#pragma implementation

#include "config.h"

#include <debtags/Update.h>
#include <debtags/PackageDB.h>
#include <debtags/TagDB.h>
#include <debtags/Vocabulary.h>
#include <debtags/Environment.h>
#include <debtags/ZlibParserInput.h>

#include <apt-pkg/acquire.h>
#include <apt-pkg/acquire-item.h>
#include <apt-pkg/strutl.h>
#include <apt-pkg/configuration.h>
#include <apt-pkg/error.h>
#include <apt-pkg/init.h>

#include <tagcoll/OpSet.h>

#include <tagcoll/TagcollParser.h>
#include <tagcoll/TagcollFilter.h>
#include <tagcoll/FilterChain.h>
#include <tagcoll/InputMerger.h>
#include <tagcoll/ItemGrouper.h>
#include <tagcoll/TagcollSerializer.h>
#include <tagcoll/UnfacetedRemover.h>

#include <string>


#include <sys/types.h>	// chmod
#include <sys/stat.h>	// chmod

#include <errno.h>

#include <stdio.h>	// For the fprintf (FIXME: remove this, and the printfs)


using namespace std;
using namespace Tagcoll;

static const char* fn_sources = "/etc/debtags/sources.list";
static const char* path_cache = "/var/cache/debtags";

//fn_dist_vocab = '$(usr/share/debtags)'/vocabulary
//path_tagpatch_d = '$(sysconfigdir)'/debtags/tagpatch.d
//path_tagvoc_d = '$(sysconfigdir)'/debtags/tagvoc.d
//remote_tags_fname = tags-current.gz
//remote_vocab_fname = vocabulary.gz

namespace Debtags
{

// TagcollFilter that removes tags not in the given Vocabulary
template<class ITEM, class TAG = std::string>
class VocabularyFilter : public TagcollFilter<ITEM, TAG>
{
protected:
	Debtags::Vocabulary& voc;
	
public:
	VocabularyFilter(Debtags::Vocabulary& voc) throw () : voc(voc) {}

	virtual void consume(const std::string& item) throw ();
	virtual void consume(const std::string& item, const OpSet<std::string>& tags) throw ();
	virtual void consume(const OpSet<ITEM>& items) throw ();
	virtual void consume(const OpSet<ITEM>& items, const OpSet<TAG>& tags) throw ();
};

template<class ITEM, class TAG>
void VocabularyFilter<ITEM, TAG>::consume(const string& item) throw ()
{
	consumer->consume(item);
}
template<class ITEM, class TAG>
void VocabularyFilter<ITEM, TAG>::consume(const OpSet<ITEM>& items) throw ()
{
	consumer->consume(items);
}


template<class ITEM, class TAG>
void VocabularyFilter<ITEM, TAG>::consume(const string& item, const OpSet<string>& tags) throw ()
{
	OpSet<string> patched;
	for (OpSet<string>::const_iterator i = tags.begin();
			i != tags.end(); i++)
		if (voc.hasTag(*i))
			patched += *i;
//		else
//			fprintf(stderr, "Removing tag %.*s not found in vocabulary (package %.*s)\n", PFSTR(*i), PFSTR(item));

	if (patched.size())
		consumer->consume(item, patched);
	else
		consumer->consume(item);
}

// Process a set of items identically tagged, with their tags
template<class ITEM, class TAG>
void VocabularyFilter<ITEM, TAG>::consume(const OpSet<ITEM>& items, const OpSet<TAG>& tags) throw ()
{
	OpSet<string> patched;
	for (OpSet<string>::const_iterator i = tags.begin();
			i != tags.end(); i++)
		if (voc.hasTag(*i))
			patched += *i;
		else
		{
			string pkglist;
			for (typename OpSet<ITEM>::const_iterator j = items.begin();
					j != items.end(); j++)
			{
				if (j == items.begin())
					pkglist += *j;
				else
					pkglist += ", " + *j;
			}
//			fprintf(stderr, "Removing tag %.*s not found in vocabulary (package%s %.*s)\n", PFSTR(*i), items.size() > 1 ? "s" : "", PFSTR(pkglist));
		}

	if (patched.size())
		consumer->consume(items, patched);
	else
		consumer->consume(items);
}

// TagcollFilter that removes items not in the APT database
template<class ITEM, class TAG = std::string>
class APTFilter : public TagcollFilter<ITEM, TAG>
{
protected:
	PackageDB packageDB;
public:
	APTFilter() throw () : packageDB(Debtags::Environment::get().packageDB()) {}
	virtual ~APTFilter() throw () {}
	virtual void consume(const std::string& item) throw ();
	virtual void consume(const std::string& item, const OpSet<std::string>& tags) throw ();
	virtual void consume(const OpSet<ITEM>& items) throw ();
	virtual void consume(const OpSet<ITEM>& items, const OpSet<TAG>& tags) throw ();
};

template<class ITEM, class TAG>
void APTFilter<ITEM, TAG>::consume(const string& item) throw ()
{
	if (packageDB.hasPackage(item))
		consumer->consume(item);
//	else
//		fprintf(stderr, "Removing package %.*s not found in APT database\n", PFSTR(item));
}
template<class ITEM, class TAG>
void APTFilter<ITEM, TAG>::consume(const OpSet<ITEM>& items) throw ()
{
	OpSet<ITEM> filtered;
	for (typename OpSet<ITEM>::const_iterator i = items.begin();
			i != items.end(); i++)
		if (packageDB.hasPackage(*i))
			filtered += *i;
//		else
//			fprintf(stderr, "Removing package %.*s not found in APT database\n", PFSTR(*i));
	consumer->consume(filtered);
}


template<class ITEM, class TAG>
void APTFilter<ITEM, TAG>::consume(const string& item, const OpSet<string>& tags) throw ()
{
	if (packageDB.hasPackage(item))
		consumer->consume(item, tags);
//	else
//		fprintf(stderr, "Removing package %.*s not found in APT database\n", PFSTR(item));
}

// Process a set of items identically tagged, with their tags
template<class ITEM, class TAG>
void APTFilter<ITEM, TAG>::consume(const OpSet<ITEM>& items, const OpSet<TAG>& tags) throw ()
{
	OpSet<ITEM> filtered;
	for (typename OpSet<ITEM>::const_iterator i = items.begin();
			i != items.end(); i++)
		if (packageDB.hasPackage(*i))
			filtered += *i;
//		else
//			fprintf(stderr, "Removing package %.*s not found in APT database\n", PFSTR(*i));
	consumer->consume(filtered, tags);
}

};


namespace DebtagsUpdate {

struct source
{
	enum type_t { TAGS };

	type_t type;
	string uri;
	source(type_t type, const std::string& uri) throw () : type(type), uri(uri) {}
};

static void zreadCollection(const std::string& local, Debtags::Vocabulary& voc, InputMerger<string, string>& merger)
	throw (FileException, ParserException)
{
	// Read data from local
	ZlibParserInput in(local);

	FilterChain<string, string> filters;

	// Filter out tags not in voc
	Debtags::VocabularyFilter<string> vocFilter(voc);
	filters.appendFilter(&vocFilter);

	// Filter out packages not in APT
	Debtags::APTFilter<string> aptFilter;
	filters.appendFilter(&aptFilter);

	filters.setConsumer(&merger);

	// Read the collection
	Tagcoll::parseTagcoll(in, filters);

	// Add packages only in APT as untagged
	// (unneeded as InputMerger already returns an empty tagset for non-existing items)
}

static void importData(const std::string& local, InputMerger<string, string>& merger, Debtags::Vocabulary& systemvoc) throw ()
{
	zreadCollection(local, systemvoc, merger);
}

static void importData(const std::string& local, const std::string& localVoc, InputMerger<string, string>& merger, Debtags::Vocabulary& systemvoc) throw ()
{
	Debtags::Vocabulary voc;
	ZlibParserInput in(localVoc);
	voc.read(in);

	zreadCollection(local, voc, merger);

	// Merge vocabularies together
	ZlibParserInput in2(localVoc);
	systemvoc.read(in2);
}

static vector<source> readSources() throw (FileException)
    // TODO: remove in favor of apt-pkg sources.list parser?
{
	FILE* in = fopen(fn_sources, "rt");
	if (!in)
		return vector <source>();
		// throw FileException(errno, string("reading ") + fn_sources);

	vector<source> res;

	string line;
	int c;
	while ((c = fgetc(in)) != EOF)
	{
		if (c != '\n')
			line += c;
		else
		{
			unsigned int i = 0;

			// Skip leading spaces
			while (i < line.size() && isspace(line[i]))
				i++;

			// If it's a tag source
			if (line.substr(i, 4) == "tags")
			{
				i += 4;
				// Skip further spaces
				while (i < line.size() && isspace(line[i]))
					i++;
				res.push_back(source(source::TAGS, line.substr(i)));
			}
			
			/*
			// If it's a debtags source
			if (line.substr(i, 4) == "debtags")
			{
				i += 7;
				// Skip further spaces
				while (i < line.size() && isspace(line[i]))
					i++;
				res.push_back(source(source::DEBTAGS, line.substr(i)));
			}
			*/
			
			line = string();
		}
	}

	fclose(in);

	return res;
}

// Item class for index files
class AcqTagfile: public pkgAcquire::Item
{
	protected:
		bool decompression;
		bool erase;
		pkgAcquire::ItemDesc Desc;
		string RealURI;

	public:

		// Specialized action members
		virtual void Done(string Message,unsigned long Size,string Md5Hash,
					  pkgAcquire::MethodConfig *Cnf);
		virtual string DescURI() {return RealURI + ".gz";};

		AcqTagfile(pkgAcquire *Owner,string URI,string URIDesc,
				string ShortDesct);
};

// AcqTagfile::AcqTagfile - Constructor                                     /*{{{*/
// ---------------------------------------------------------------------
/* The package file is added to the queue and a second class is
   instantiated to fetch the revision file */
AcqTagfile::AcqTagfile(pkgAcquire *Owner,
                         string URI,string URIDesc,string ShortDesc) :
   Item(Owner), RealURI(URI)
{
   //fprintf(stderr, "AcqTagfile::AcqTagfile ('%.*s') called\n", PFSTR(URI));
   decompression = false;
   erase = false;

   DestFile = _config->FindDir("Dir::State::lists") + "partial/";
   DestFile += URItoFileName(URI);
   //fprintf(stderr, "DestFile: %.*s\n", PFSTR(DestFile));

   // Create the item
   Desc.URI = URI;
   Desc.Description = URIDesc;
   Desc.Owner = this;
   Desc.ShortDesc = ShortDesc;

   QueueURI(Desc);
}
                                                                        /*}}}*/
// AcqTagfile::Done - Finished a fetch                                    /*{{{*/
// ---------------------------------------------------------------------
/* This goes through a number of states.. On the initial fetch the
   method could possibly return an alternate filename which points
   to the uncompressed version of the file. If this is so the file
   is copied into the partial directory. In all other cases the file
   is decompressed with a gzip uri. */
void AcqTagfile::Done(string Message,unsigned long Size,string MD5,
                       pkgAcquire::MethodConfig *Cfg)
{
   Item::Done(Message,Size,MD5,Cfg);
   //fprintf(stderr, "AcqTagfile::Done (%.*s) called.\n", PFSTR(Message));

   if (decompression == true)
   {
      /* if (MD5.empty())
      {
         MD5Summation sum;
         FileFd Fd(DestFile, FileFd::ReadOnly);
         sum.AddFD(Fd.Fd(), Fd.Size());
         Fd.Close();
         MD5 = (string)sum.Result();
      } */

      // Done, move it into position
      //string FinalFile = _config->FindDir("Dir::State::lists") + "tags/";
      string FinalFile = path_cache;
      FinalFile += "/" + URItoFileName(RealURI);
	  //fprintf(stderr, "FinalFile: %.*s\n", PFSTR(FinalFile));
      Rename(DestFile,FinalFile);
      chmod(FinalFile.c_str(),0644);

      /* We restore the original name to DestFile so that the clean operation
         will work OK */
	  DestFile = _config->FindDir("Dir::State::lists") + "partial/";
	  DestFile += URItoFileName(RealURI);
	  //fprintf(stderr, "DestFile: %.*s\n", PFSTR(DestFile));

      // Remove the compressed version.
      if (erase == true)
         unlink(DestFile.c_str());
      return;
   }

   erase = false;
   Complete = true;

   // Handle the unzipd case
   string FileName = LookupTag(Message,"Alt-Filename");
   if (FileName.empty() == false)
   {
      // The files timestamp matches
      if (StringToBool(LookupTag(Message,"Alt-IMS-Hit"),false) == true)
         return;

      decompression = true;
      Local = true;
      DestFile += ".decomp";
      Desc.URI = "copy:" + FileName;
      QueueURI(Desc);
      Mode = "copy";
      return;
   }
   FileName = LookupTag(Message,"Filename");
   if (FileName.empty() == true)
   {
      Status = StatError;
      ErrorText = "Method gave a blank filename";
   }

   // The files timestamp matches
   if (StringToBool(LookupTag(Message,"IMS-Hit"),false) == true)
      return;

   if (FileName == DestFile)
      erase = true;
   else
      Local = true;

   decompression = true;
   DestFile += ".decomp";
   Desc.URI = "gzip:" + FileName;
   QueueURI(Desc);
   Mode = "gzip";
}

class basicAcquireStatus : public pkgAcquireStatus
{
public:
	virtual bool MediaChange(string Media,string Drive) { return true; }
	virtual void IMSHit(pkgAcquire::ItemDesc &/*Itm*/)
	{
		fprintf(stderr, "IMSHit\n");
	}
	virtual void Fetch(pkgAcquire::ItemDesc &/*Itm*/)
	{
		//fprintf(stderr, "Fetch\n");
	}
	virtual void Done(pkgAcquire::ItemDesc &/*Itm*/)
	{
		//fprintf(stderr, "Done\n");
	}
	virtual void Fail(pkgAcquire::ItemDesc &Itm)
	{
		fprintf(stderr, "Failed downloading from %.*s\n", PFSTR(Itm.URI));
		_error->DumpErrors();
	}
	/*
	virtual bool Pulse(pkgAcquire *Owner)
	{
		fprintf(stderr, "PULSE\n");
		return pkgAcquireStatus::Pulse(Owner);
	}
	virtual void Start()
	{
		fprintf(stderr, "START\n");
		pkgAcquireStatus::Start();
	}
	virtual void Stop()
	{
		fprintf(stderr, "STOP\n");
		pkgAcquireStatus::Stop();
	}
	*/
};

void updateDebtagsDatabase()
	throw (ConsistencyCheckException)
{
	basicAcquireStatus status;
	updateDebtagsDatabase(&status);
}

void updateDebtagsDatabase(pkgAcquireStatus* status)
	throw (ConsistencyCheckException)
{
	if (!_config->Exists("Dir::State"))
	{
		// Need to read the libapt-pkg configuration...
		pkgInitConfig(*_config);
	}
	
	vector<source> sources = readSources();

	pkgAcquire fetcher(status);

	vector<AcqTagfile*> acquirers;
	// TODO: deallocate acquirers!
	for (vector<source>::const_iterator i = sources.begin();
			i != sources.end(); i++)
		switch (i->type)
		{
			case source::TAGS:
				acquirers.push_back(new AcqTagfile (&fetcher, i->uri + "tags-current.gz", i->uri + "tags-current.gz", "Tag database"));
				acquirers.push_back(new AcqTagfile (&fetcher, i->uri + "vocabulary.gz", i->uri + "vocabulary.gz", "Tag vocabulary"));
				break;
			/*
			case source::DEBTAGS:
				acquirers.push_back(new AcqTagfile (&fetcher, i->uri + "tags-current.gz", "", ""));
				acquirers.push_back(new AcqTagfile (&fetcher, i->uri + "vocabulary.gz", "", ""));
				break;
				*/
		}

	if (fetcher.Run() == pkgAcquire::Failed)
	{
		_error->DumpErrors();
		// FIXME: need a better message as soon as I understand what's going on
		throw ConsistencyCheckException("Acquirer failed");
	}

	for (pkgAcquire::ItemIterator it = fetcher.ItemsBegin (); it != fetcher.ItemsEnd (); it ++)
	{
		if ((*it)->Status == pkgAcquire::Item::StatDone)
			continue;
		(*it)->Finished();
		// TODO: warn user about failures
		// TODO: I would like to: how?
	}

	InputMerger<string, string> merger;
	Debtags::Vocabulary voc;
	voc.readVocabularies();
	int found = 0;
	for (vector<source>::const_iterator i = sources.begin();
			i != sources.end(); i++)
	{
		//fprintf(stderr, "Acquiring %s source %.*s...\n", i->type == source::TAGS ? "tags" : "debtags",
				//PFSTR(i->uri));

		try {
			//string local = _config->FindDir("Dir::State::lists") + "tags/";
			string local = path_cache;
			local += "/" + URItoFileName(i->uri + "tags-current.gz");
			if (access(local.c_str(), R_OK) == -1)
			{
				fprintf(stderr, "Source %.*s has problems: ignoring it\n", PFSTR(i->uri));
				continue;
			}

			//string localVoc = _config->FindDir("Dir::State::lists") + "tags/";
			string localVoc = path_cache;
			localVoc += "/" + URItoFileName(i->uri + "vocabulary.gz");

			if (access(localVoc.c_str(), R_OK) == -1)
			{
				fprintf(stderr, "Reading tag data for %.*s...\n", PFSTR(i->uri));
				importData(local, merger, voc);
			} else {
				fprintf(stderr, "Reading tag data and vocabulary for %.*s...\n", PFSTR(i->uri));
				importData(local, localVoc, merger, voc);
			}

			found++;
		} catch (Exception& e) {
			fprintf(stderr, "%s: %.*s.  Ignoring source %.*s\n", e.type(), PFSTR(e.desc()), PFSTR(i->uri));
		}
	}

	if (!found)
	{
		_error->DumpErrors();
		throw ConsistencyCheckException("Unable to use any data source (not even previously cached ones)");
	}

	// Write merged vocabulary
	fprintf(stderr, "Writing system vocabulary...\n");
	voc.write();
	//voc.writeImplications();
	//voc.writeDerivedTags();

	// Write merged tag database
	fprintf(stderr, "Writing merged tag database...\n");
	// Group the items before output
	ItemGrouper<string, string> grouper;
	merger.output(grouper);
	
	// Output the merged collection to the target file
	string tagdb = Debtags::TagDB::path_tagdb();
	string tmpdb = tagdb + ".tmp";
	FILE* out = fopen(tmpdb.c_str(), "wt");
	if (!out) throw FileException(errno, "opening " + tmpdb);
	TagcollSerializer serializer(out);
	grouper.output(serializer);
	fclose(out);

	// Perform atomic update of the tag database
	if (rename(tmpdb.c_str(), tagdb.c_str()) == -1)
		throw FileException(errno, "renaming " + tmpdb + " to " + tagdb);
}

/*
				// Prepare the input filter chain
				FilterChain<string> filters;

				// Build the patch list
				PatchList patches;
				int patchCount = 0;
				DIR* dir = opendir(path_tagpatch_d);
				if (!dir)
					throw SystemException(errno, string("reading directory ") + path_tagpatch_d);
				while (struct dirent* d = readdir(dir))
				{
					if (d->d_name[0] != '.' &&
							d->d_name[strlen(d->d_name) - 1] != '~')
					{
						string fn = string(path_tagpatch_d) + '/' + d->d_name;
						if (access(fn.c_str(), R_OK) == 0)
						{
							readCollection(fn, patches);
							patchCount++;
						}
					}
				}
				closedir(dir);


				// Prepare applying the patches
				if (patchCount)
					filters.appendFilter(new ApplyPatches(patches));

				ImplicationList implications;
				DerivedTagList derivedTags;

				voc.outputImplications(implications);
				voc.outputDerivedTags(derivedTags);

				implications.pack();

				// Prepare packing the database data

				// Expand implications
				filters.appendFilter(new ApplyImplications(implications));
				// Remove derived tags computing them using the expanded tag set
				filters.appendFilter(new CompressDerivedTags(derivedTags));
				// Compress implications
				filters.appendFilter(new CompressImplications(implications));
*/

};

// vim:set ts=4 sw=4:
