/*
 * Tag vocabulary access
 *
 * Copyright (C) 2003  Enrico Zini <enrico@debian.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <debtags/Vocabulary.h>
#include <debtags/Environment.h>

#include <tagcoll/tagexpr/Tagexpr.h>
#include <tagcoll/tagexpr/TagexprParser.h>
#include <tagcoll/StdioParserInput.h>
#include <tagcoll/StringParserInput.h>
#include <tagcoll/TagcollParser.h>
#include <tagcoll/stringf.h>

#include <apt-pkg/tagfile.h>

#include <sys/types.h>  // opendir, readdir, closedir
#include <dirent.h>     // opendir, readdir, closedir
#include <unistd.h>     // access
#include <errno.h>

using namespace std;
using namespace Debtags;

static const char* fn_dist_vocab = "/usr/share/debtags/vocabulary";
static const char* path_tagvoc_d = "/etc/debtags/tagvoc.d";

namespace {
	
class VocabularyParser
{
protected:
	string buffer;
	ParserInput& input;

public:
	VocabularyParser(ParserInput& input) throw ()
		: input(input) {}

	bool getNextRecord(pkgTagSection& section) throw (ParserException)
	{
		buffer.resize(0);
		int nlines = 0;
		while (true)
		{
			int c = input.nextChar();
			switch (c)
			{
				case ParserInput::Eof:
					if (buffer.empty())
						return false;
					else
					{
						section.Scan(buffer.data(), buffer.size());
						return true;
					}
				case '\n':
					if (buffer.empty())
						break;
					nlines++;
					buffer += c;
					if (nlines == 2)
					{
						nlines = 0;
						section.Scan(buffer.data(), buffer.size());
						return true;
					}
					break;
				default:
					nlines = 0;
					buffer += c;
					break;
			}
		}
	}	
};

};

static void writeDebStyleField(FILE* out, const string& name, const string& val) throw ()
{
	fprintf(out, "%.*s: ", PFSTR(name));

	// Properly escape newlines
	bool was_nl = false;
	for (string::const_iterator s = val.begin(); s != val.end(); s++)
		if (was_nl)
			// \n\n -> \n .\n
			if (*s == '\n')
			{
				fputc(' ', out);
				fputc('.', out);
				fputc(*s, out);
			}
			// \n([^ \t]) -> \n \1
			else if (*s != ' ' && *s != '\t')
			{
				fputc(' ', out);
				fputc(*s, out);
				was_nl = false;
			}
			// \n[ \t] goes unchanged
			else
			{
				fputc(*s, out);
				was_nl = false;
			}
		else
			if (*s == '\n')
			{
				fputc(*s, out);
				was_nl = true;
			}
			else
				fputc(*s, out);

	fputc('\n', out);
}

void Vocabulary::read(ParserInput& input) throw (ParserException)
{
	VocabularyParser parser(input);
	pkgTagSection section;

	while (parser.getNextRecord(section))
	{
		string tname = section.FindS("Tag");
		string fname = section.FindS("Facet");
		if (!tname.empty())
		{
			Facet facet;
			Tag tag;
			
			unsigned int p = tname.find("::");
			if (p == string::npos)
			{
				facet = obtainFacet("");
				tag = facet.obtainTag(tname);
				if (!tag)
					fprintf(stderr, "Bug loading vocabulary! obtained tag is false\n");
				if (tag.fullname() != tname)
					fprintf(stderr, "Bug loading vocabulary! [%.*s] became [%.*s]\n", PFSTR(tname), PFSTR(tag.fullname()));
			} else {
				facet = obtainFacet(tname.substr(0, p));
				tag = facet.obtainTag(tname.substr(p + 2));
				if (!tag)
					fprintf(stderr, "Bug loading vocabulary! obtained tag is false\n");
				if (tag.fullname() != tname)
					fprintf(stderr, "Bug loading vocabulary! [%.*s] became [%.*s]\n", PFSTR(tname), PFSTR(tag.fullname()));
			}

			string expr = section.FindS("Equates");
			if (!expr.empty())
			{
				Tagexpr* compiledExpr = TagexprParser::instance()->parse(expr);
				if (compiledExpr)
					equations.add(tname, compiledExpr);
				else
				{
					string err = TagexprParser::instance()->getErrorMessage();
					fprintf(stderr, "Cannot parse Equates in tag %.*s: %.*s.  Ignoring expression.\n", PFSTR(tag.fullname()), PFSTR(err));
				}
			}

			string desc = section.FindS("Description");
			if (!desc.empty())
			{
				int send = desc.find('\n');
				tag.sdesc(desc.substr(0, send));
				tag.ldesc(desc.substr(send + 1));
			}

			const char* start;
			const char* stop;
			section.GetSection(start, stop);
			tag.fulldata(string(start, stop-start));
		}
		else if (!fname.empty())
		{
			Facet facet = obtainFacet(fname);
			if (!facet)
				fprintf(stderr, "Bug loading vocabulary! obtained facet is false\n");
			if (facet.name() != fname)
				fprintf(stderr, "Bug loading vocabulary! [%.*s] became [%.*s]\n", PFSTR(fname), PFSTR(facet.name()));

			string desc = section.FindS("Description");
			if (!desc.empty())
			{
				int send = desc.find('\n');
				facet.sdesc(desc.substr(0, send));
				facet.ldesc(desc.substr(send + 1));
			}

			const char* start;
			const char* stop;
			section.GetSection(start, stop);
			facet.fulldata(string(start, stop-start));
		}
		else
		{
			fprintf(stderr, "%.*s:%d: Skipping record without Tag or Facet field\n",
					PFSTR(input.fileName()), input.lineNumber());
		}
	}
}

void Vocabulary::read() throw (FileException, ParserException)
{
	StdioParserInput mainInput(Debtags::Environment::get().path_vocabulary());
	read(mainInput);
}

void Vocabulary::readVocabularies() throw (SystemException, FileException, ParserException)
{
	StdioParserInput mainInput(fn_dist_vocab);
	read(mainInput);
	
	DIR* dir = opendir(path_tagvoc_d);
	if (!dir)
		throw SystemException(errno, string("reading directory ") + path_tagvoc_d);

	while (struct dirent* d = readdir(dir))
	{
		if (d->d_name[0] != '.' &&
				d->d_name[strlen(d->d_name) - 1] != '~')
		{
			string fn = string(path_tagvoc_d) + '/' + d->d_name;
			if (access(fn.c_str(), R_OK) == 0)
			{
				StdioParserInput patchInput(fn);
				read(patchInput);
			}
		}
	}
	closedir(dir);
}

void Vocabulary::write() throw (FileException)
{
	string fn_vocab = Debtags::Environment::get().path_vocabulary();
	string fn = string(fn_vocab) + ".tmp";
	FILE* out = fopen(fn.c_str(), "wt");
	if (!out) throw FileException(errno, "creating file " + fn);

	for (FacetSet::const_iterator i = begin(); i != end(); i++)
	{
		string buf = i->fulldata();
		if (buf.empty())
		{
			if (!i->name().empty())
			{
				writeDebStyleField(out, "Facet", i->name());
				fputc('\n', out);
			}
		} else
			fwrite(buf.data(), buf.size(), 1, out);

		for (TagSet::const_iterator j = i->tags().begin(); j != i->tags().end(); j++)
		{
			string buf = j->fulldata();
			if (buf.empty())
			{
				writeDebStyleField(out, "Tag", i->name());
				fputc('\n', out);
			} else
				fwrite(buf.data(), buf.size(), 1, out);

			/*
			writeDebStyleField(out, "Tag", j->fullname());
			if (!j->sdesc().empty())
				if (!j->ldesc().empty())
					writeDebStyleField(out, "Description", j->sdesc() + "\n" + j->ldesc());
				else
					writeDebStyleField(out, "Description", j->sdesc());
			*/
			// TODO
			/*
			for (list< pair<string, string> >::const_iterator j = i->second.others.begin();
					j != i->second.others.end(); j++)
				if (j->second.size())
					writeDebStyleField(out, j->first, j->second);
			*/
			//fputc('\n', out);
		}
	}
	/*
	for (const_iterator i = begin(); i != end(); i++)
	{
		writeDebStyleField(out, "Tag", i->first);
		if (i->second.implies.size())
			writeDebStyleField(out, "Implies", i->second.formatImplies());
		if (i->second.derived)
		{
			Tagexpr* e = i->second.derived;
			writeDebStyleField(out, "Derived", e->format());
			//writeDebStyleField(out, "Derived", i->second.derived->format());
		}
		if (i->second.description.size())
			writeDebStyleField(out, "Description", i->second.description);

		for (list< pair<string, string> >::const_iterator j = i->second.others.begin();
				j != i->second.others.end(); j++)
			if (j->second.size())
				writeDebStyleField(out, j->first, j->second);

		fputc('\n', out);
	}
	*/

	fclose(out);

	if (rename(fn.c_str(), fn_vocab.c_str()) == -1)
		throw FileException(errno, "renaming " + fn + " to " + fn_vocab);
}

/*
void Vocabulary::outputImplications(TagcollConsumer<string>& consumer) const throw ()
{
	for (const_iterator i = begin(); i != end(); i++)
		if (!i->second.implies.empty())
		{
			OpSet<string> ts;
			copy(i->second.implies.begin(), i->second.implies.end(), inserter(ts, ts.begin()));
			consumer.consume(i->first, ts);
		}
}

void Vocabulary::outputDerivedTags(DerivedTagList& dtlist) const throw ()
{
	for (const_iterator i = begin(); i != end(); i++)
		if (i->second.derived)
			dtlist.add(i->first, i->second.derived);
}
*/

// vim:set ts=4 sw=4:
